from unittest import TestCase

from nose.plugins.attrib import attr

from findd.cli import main


@attr('integration')
class Main(TestCase):
    def test_should_throw_exceptions_in_debug_mode(self):
        with self.assertRaises(Exception):
            main(['update', '-vv'])



from blinker import signal


class Progress(object):

    def __init__(self, task):
        self.progress_signal = signal('progress')
        self.task = task

    def start(self, sender, maxval=None):
        self.progress_signal.send(
            sender,
            task=self.task,
            state='start',
            maxval=maxval,
        )

    def update(self, sender, val=None):
        self.progress_signal.send(
            sender,
            task=self.task,
            state='update',
            val=val,
        )

    def finish(self, sender):
        self.progress_signal.send(
            sender,
            task=self.task,
            state='finish',
        )



import itertools
import time

from backports.shutil_get_terminal_size import get_terminal_size
from blinker import signal
import progressbar


DEBOUNCE_THRESHOLD = 0.05


def hr(text='', fillchar='*'):
    cols = get_terminal_size((80, 20)).columns
    return text.center(cols, fillchar)


class ProgressBarManager(object):

    def __init__(self):
        self.bars = {}
        self.last_update = {}

    def __enter__(self):
        signal('progress').connect(self.process_progress_signal)
        return self

    def __exit__(self, *_):
        signal('progress').disconnect(self.process_progress_signal)

    def process_progress_signal(self, _, task, state, **kw):
        if state == 'start':
            speed = eta = task == 'hashing'
            maxval = kw['maxval'] if 'maxval' in kw else None
            self.start(task, maxval=maxval, eta=eta, speed=speed)
        elif state == 'update':
            self.update(task, val=kw['val'] if 'val' in kw else None)
        elif state == 'finish':
            self.finish(task)

    def start(self, task, maxval=None, eta=False, speed=False):
        label_width = 16
        widgets = []
        if maxval is None:
            widgets.append(task.ljust(label_width))
            widgets.append(' ')
            widgets.append(progressbar.BouncingBar())
        else:
            widgets.append(task.ljust(label_width - 5))
            widgets.append(' ')
            widgets.append(progressbar.Percentage())
            widgets.append(' ')
            widgets.append(progressbar.Bar())
        if eta and maxval != 0:
            widgets.append(' ')
            widgets.append(progressbar.ETA())
        if speed and maxval != 0:
            widgets.append(' ')
            widgets.append(progressbar.FileTransferSpeed())
        pbar = progressbar.ProgressBar(
            widgets=widgets,
            maxval=maxval if maxval is None or maxval > 0 else 1,
        )
        if maxval is None:
            self.bars[task] = iter(pbar(itertools.repeat(0)))
        else:
            self.bars[task] = pbar.start()
        self.last_update[task] = 0

    def update(self, task, val=None):
        if time.time() - self.last_update[task] < DEBOUNCE_THRESHOLD:
            return  # pragma: no cover
        if val is None:
            self.bars[task].next()
        else:
            self.bars[task].update(val)
        self.last_update[task] = time.time()

    def finish(self, task):
        self.bars[task].finish()
        del self.bars[task]
        del self.last_update[task]



import os
import subprocess

from findd.cli import context
from findd.cli.views import InitView
from findd.cli.views import ListDuplicatesView
from findd.cli.views import ProcessDuplicatesView
from findd.cli.views import UpdateView


def init(_):
    with InitView():
        ctx = context.create_from_base_dir(base_dir=os.getcwd())
        ctx.assert_findd_dir_does_not_exists()
        os.makedirs(ctx.findd_dir)
        ctx.findd().init()


def update(opts):
    with UpdateView():
        ctx = context.create_from_base_dir()
        ctx.assert_findd_dir_exists()
        ctx.findd().update(is_excluded=ctx.is_excluded)


def list_duplicates(opts):
    with ListDuplicatesView() as view:
        ctx = context.create_from_base_dir()
        ctx.assert_findd_dir_exists()
        for dups in ctx.findd().duplicates(limit=opts.limit):
            view.print_duplicates(dups)


def process_duplicates(opts):
    with ProcessDuplicatesView() as view:
        ctx = context.create_from_base_dir()
        ctx.assert_findd_dir_exists()
        for dups in ctx.findd().duplicates(limit=opts.limit):
            paths = [afile.relpath for afile in dups]
            call_args = [opts.cmd] + opts.args + paths
            view.print_subprocess_call(call_args)
            subprocess.call(call_args)
            view.print_duplicates(dups)



from io import BytesIO
from mock import patch
from unittest import TestCase

from findd.utils.crypto import hashfile


class Hashfile(TestCase):

    @patch('findd.utils.crypto.open', create=True)
    def test_should_hash_empty_files(self, mopen):
        mopen.return_value = BytesIO()
        h = hashfile('empty')
        mopen.assert_called_once_with('empty', 'rb')

        self.assertEqual(
            h['md5'],
            'd41d8cd98f00b204e9800998ecf8427e'
        )
        self.assertEqual(
            h['sha1'],
            'da39a3ee5e6b4b0d3255bfef95601890afd80709'
        )
        self.assertEqual(
            h['sha224'],
            'd14a028c2a3a2bc9476102bb288234c415a2b01f828ea62ac5b3e42f'
        )
        self.assertEqual(
            h['sha256'],
            'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
        )
        self.assertEqual(
            h['sha384'],
            '38b060a751ac96384cd9327eb1b1e36a21fdb71114be07434c0cc7bf63f6e1da274edebfe76f65fbd51ad2f14898b95b'  # noqa
        )
        self.assertEqual(
            h['sha512'],
            'cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e'  # noqa
        )

    @patch('findd.utils.crypto.open', create=True)
    def test_should_hash_abc_files(self, mopen):
        mopen.return_value = BytesIO(b'abc')
        h = hashfile('abc')
        mopen.assert_called_once_with('abc', 'rb')

        self.assertEqual(
            h['md5'],
            '900150983cd24fb0d6963f7d28e17f72'
        )

        self.assertEqual(
            h['sha1'],
            'a9993e364706816aba3e25717850c26c9cd0d89d'
        )
        self.assertEqual(
            h['sha224'],
            '23097d223405d8228642a477bda255b32aadbce4bda0b3f7e36c9da7'
        )
        self.assertEqual(
            h['sha256'],
            'ba7816bf8f01cfea414140de5dae2223b00361a396177a9cb410ff61f20015ad'
        )
        self.assertEqual(
            h['sha384'],
            'cb00753f45a35e8bb5a03d699ac65007272c32ab0eded1631a8b605a43ff5bed8086072ba1e7cc2358baeca134c825a7'  # noqa
        )
        self.assertEqual(
            h['sha512'],
            'ddaf35a193617abacc417349ae20413112e6fa4e89a97ea20a9eeee64b55d39a2192992a274fc1a836ba3c23a3feebbd454d4423643ce80e2a9ac94fa54ca49f'  # noqa
        )



__version__ = '0.8.8'
__url__ = 'https://github.com/schnittstabil/findd'
__bugs_url__ = __url__ + '/issues'
__author__ = 'Michael Mayer'
__author_email__ = 'michael@schnittstabil.de'
__license__ = 'MIT'



from mock import patch
from unittest import TestCase

from findd.services import HashQueue


class HashQueueTest(TestCase):

    @patch('findd.services.hashfile')
    def test_should_not_raise_on_ioerrors(self, hashfile):
        sut = HashQueue()
        sut.append('a.txt', 1)
        sut.append('b.txt', 1)
        sut.append('c.txt', 1)

        hashfile.side_effect = [
            {'md5': 'a'},
            IOError(),
            {'md5': 'c'},
        ]

        results = sut.process()

        self.assertEqual(len(hashfile.mock_calls), 3)
        self.assertEqual(results, [
            {'md5': 'a'},
            {'md5': 'c'},
        ])

        self.assertEqual(len(sut.errors), 1)
        self.assertIsInstance(sut.errors[0], IOError)



import errno
import logging
import os
from os.path import abspath
from os.path import join
from os.path import relpath

from shellescape import quote

from findd import model
from findd.utils.crypto import hashfile
from findd.utils.format import sizeof_fmt
from findd.utils.path import files_of_dir
from findd.utils.progress import Progress
from findd.utils.unicode import u, s


__LOG__ = logging.getLogger(__name__)


def dir_aware_file_registry(base_dir, file_registry):
    model_find_all = file_registry.find_all
    model_find_duplicates = file_registry.find_duplicates

    def set_paths(afile):
        afile.base_dir = base_dir
        afile.abspath = abspath(join(afile.base_dir, afile.path))
        afile.relpath = relpath(afile.abspath)
        return afile

    def find_all_(*args, **kwargs):
        for afile in model_find_all(*args, **kwargs):
            yield set_paths(afile)

    def find_duplicates_(*args, **kwargs):
        for dups in model_find_duplicates(*args, **kwargs):
            yield [set_paths(afile) for afile in dups]

    file_registry.find_all = find_all_
    file_registry.find_duplicates = find_duplicates_
    return file_registry


class Findd(object):

    def __init__(self, base_dir, db_session):
        self.base_dir = base_dir
        self.db_session = db_session
        self._file_registry = None

    def init(self):
        progress = Progress('init')
        __LOG__.debug('initializing...')
        progress.start(self, maxval=1)
        model.create_schema(self.db_session.get_bind())
        progress.finish(self)

    def update(self, is_excluded=None):
        update_cmd = _UpdateCommand(
            base_dir=self.base_dir,
            file_registry=self.file_registry,
            is_excluded=is_excluded
        )
        count = update_cmd.execute()
        self.db_session.commit()
        return count

    def duplicates(self, limit=-1):
        progress = Progress('find duplicates')
        index = 0
        progress.start(self, maxval=limit if limit > 0 else None)
        for i in self.file_registry.find_duplicates(limit=limit):
            yield i
            index = index + 1
            progress.update(self, val=index)
        progress.finish(self)

    @property
    def file_registry(self):
        if self._file_registry is None:
            self._file_registry = dir_aware_file_registry(
                self.base_dir, model.FileRegistry(self.db_session)
            )
        return self._file_registry


class HashQueue(object):

    def __init__(self):
        self._queue = []
        self.bytes_to_hash = 0
        self.errors = []
        self.processed = []
        self.bytes_processed = 0

    def append(self, file_path, file_size, dest=None):
        self._queue.append((file_path, file_size, dest or {}))
        self.bytes_to_hash = self.bytes_to_hash + file_size

    def process(self):
        progress = Progress('hashing')

        __LOG__.debug(
            'hashing %d files (%s)',
            len(self._queue),
            sizeof_fmt(self.bytes_to_hash)
        )
        progress.start(self, maxval=self.bytes_to_hash)
        for (file_path, file_size, dest) in self._queue:
            __LOG__.debug(
                'hashing %s (%s)...',
                quote(file_path),
                sizeof_fmt(file_size)
            )
            try:
                hash_values = hashfile(file_path)
            except Exception as err:
                __LOG__.exception('hashing of %s failed: ', file_path)
                self.errors.append(err)
            else:
                dest.update(hash_values)
                self.processed.append(dest)
            self.bytes_processed = self.bytes_processed + file_size
            progress.update(self, val=self.bytes_processed)
        progress.finish(self)
        __LOG__.debug('%d files hashed', len(self.processed))
        return self.processed


class _UpdateCommand(object):

    def __init__(self, base_dir, file_registry, is_excluded=None):
        assert os.path.exists(base_dir)
        self.base_dir = base_dir
        self.file_registry = file_registry
        self.is_excluded = is_excluded
        self.visited_files = []
        self.hash_queue = HashQueue()

    def _step0_scan_db(self):
        progress = Progress('scanning db')
        __LOG__.debug('scanning db...')
        db_entries_count = self.file_registry.count()
        old_files_deleted = 0
        mtime_changed = 0
        size_changed = 0
        index = 0
        progress.start(self, maxval=db_entries_count)
        for db_file in self.file_registry.find_all():
            index = index + 1
            progress.update(self, val=index)
            abs_path = join(self.base_dir, db_file.path)
            try:
                changed = False
                stat = os.stat(abs_path)
                if db_file.mtime != int(stat.st_mtime):
                    mtime_changed = mtime_changed + 1
                    changed = True
                if db_file.size != stat.st_size:
                    size_changed = size_changed + 1
                    changed = True
                if changed:
                    db_file.mtime = int(stat.st_mtime)
                    db_file.size = stat.st_size
                    self.hash_queue.append(abs_path, stat.st_size, db_file)
                self.visited_files.append(s(db_file.path))
            except OSError as err:
                if err.errno == errno.ENOENT:
                    __LOG__.debug('deleting %s', quote(db_file.path))
                    self.file_registry.delete(db_file)
                    old_files_deleted = old_files_deleted + 1
                else:
                    raise  # pragma: no cover
        progress.finish(self)
        __LOG__.debug('mtime of %d files changed', mtime_changed)
        __LOG__.debug('size of %d files changed', size_changed)
        __LOG__.debug('%d old files deleted', old_files_deleted)
        return len(self.visited_files)

    def _step1_scan_fs(self):
        progress = Progress('scanning fs')
        __LOG__.debug('scanning %s', quote(self.base_dir))
        new_files_found = 0
        progress.start(self)
        for entry in files_of_dir(self.base_dir, self.is_excluded):
            progress.update(self)
            rel_path = os.path.relpath(entry.path, self.base_dir)
            if rel_path not in self.visited_files:
                new_files_found = new_files_found + 1
                db_file = model.File(
                    path=u(rel_path),
                    mtime=int(entry.stats.st_mtime),
                    size=entry.stats.st_size,
                )
                self.hash_queue.append(
                    entry.path,
                    entry.stats.st_size,
                    db_file
                )
        progress.finish(self)
        __LOG__.debug('%d new files found', new_files_found)
        return new_files_found

    def execute(self):
        processed_files = self._step0_scan_db() + self._step1_scan_fs()
        for db_file in self.hash_queue.process():
            self.file_registry.add(db_file)
        self.visited_files = []
        __LOG__.debug('%d files processed', processed_files)
        return processed_files



from collections import namedtuple
from os import stat
from os import walk
from os.path import dirname
from os.path import join


FileEntry = namedtuple('File', 'path stats')


def parents(directory):
    last = None
    while last != directory:
        yield directory
        last = directory
        directory = dirname(directory)


def files_of_dir(directory, is_excluded=None):
    for root, dirs, files in walk(directory):
        if is_excluded is not None:
            for names in [dirs, files]:
                for name in names:
                    if is_excluded(join(root, name)):
                        names.remove(name)

        for fname in files:
            path = join(root, fname)
            yield FileEntry(path, stat(path))



import logging
from shellescape import quote
import sys

from findd.cli.widgets import hr
from findd.cli.widgets import ProgressBarManager


__LOG__ = logging.getLogger(__name__)


if sys.version_info < (3,):
    def format_path(path):
        return quote(path).encode('utf-8')
else:
    def format_path(path):
        return quote(path)


class BaseView(object):
    def __init__(self, show_progressbars):
        self.pbm = ProgressBarManager() if show_progressbars else None

    def __enter__(self):
        if self.pbm is not None:
            self.pbm.__enter__()
        return self

    def __exit__(self, *args):
        if self.pbm is not None:
            self.pbm.__exit__(*args)


class InitView(BaseView):
    def __init__(self):
        BaseView.__init__(self, __LOG__.isEnabledFor(logging.INFO))


class UpdateView(BaseView):
    def __init__(self):
        BaseView.__init__(self, __LOG__.isEnabledFor(logging.INFO))


class ListDuplicatesView(BaseView):
    def __init__(self):
        BaseView.__init__(self, False)

    def print_duplicates(self, duplicates):
        print(' '.join([format_path(afile.relpath) for afile in duplicates]))


class ProcessDuplicatesView(BaseView):
    def __init__(self):
        BaseView.__init__(self, __LOG__.isEnabledFor(logging.INFO))

    def print_subprocess_call(self, args):
        __LOG__.debug(' '.join(args))

    def print_duplicates(self, duplicates):
        if __LOG__.isEnabledFor(logging.INFO):
            print(hr(' processed duplicates '))
            paths = [format_path(afile.relpath) for afile in duplicates]
            for path in paths:
                print(path)
            print(hr())



from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
import os

from findd import __about__
from findd.cli import commands


def build():
    builder = ParserBuilder(
        prog='findd',
        homepage=__about__.__url__,
        bugs_url=__about__.__bugs_url__,
    )
    builder.add_version_argument(__about__.__version__)
    builder.add_init_parser()
    builder.add_update_parser()
    builder.add_list_parser()
    builder.add_run_parser()
    return builder.build()


class FinddArgumentParser(ArgumentParser):
    def parse_known_args(self, args=None, namespace=None):
        args, argv = ArgumentParser.parse_known_args(
            self,
            args=args,
            namespace=namespace
        )
        quite_count = getattr(args, 'quite', 0)
        verbose_count = getattr(args, 'verbose', 0)
        args.verbosity = verbose_count - quite_count
        return (args, argv)


class ParserBuilder(object):

    def __init__(self, prog, homepage, bugs_url, *args, **kwargs):
        self.prog = prog
        kwargs.setdefault(
            'epilog',
            os.linesep.join([
                'Report {prog} bugs to <{bugs_url}>',
                '{prog} home page: <{homepage}>'
            ]).format(prog=prog, homepage=homepage, bugs_url=bugs_url)
        )
        kwargs.setdefault('formatter_class', RawDescriptionHelpFormatter)
        self.parser = FinddArgumentParser(*args, prog=prog, **kwargs)
        self.subparsers = self.parser.add_subparsers()

    def add_version_argument(self, version):
        self.parser.add_argument(
            '--version',
            action='version',
            version=version
        )

    def add_init_parser(self):
        parser = self._add_subparser(
            'init',
            help='create a new Findd project',
        )
        parser.set_defaults(func=commands.init)

    def add_update_parser(self):
        parser = self._add_subparser(
            'update',
            help='update the index',
        )
        parser.set_defaults(func=commands.update)

    def add_list_parser(self):
        parser = self._add_subparser(
            'list',
            help='list duplicates tracked by the index',
        )
        parser.set_defaults(func=commands.list_duplicates)
        ParserBuilder._add_limit_argument(parser)

    def add_run_parser(self):
        parser = self._add_subparser(
            'run',
            help='run a command for duplicates tracked by the index',
            epilog=os.linesep.join([
                'examples:',
                '  {prog} run -- mplayer',
                '  {prog} run --limit=1 -- ls -lh',
            ]).format(prog=self.prog),
            formatter_class=RawDescriptionHelpFormatter,
        )
        parser.set_defaults(func=commands.process_duplicates)
        parser.add_argument(
            'cmd',
            help='command to run for each set of duplicates',
        )
        parser.add_argument(
            'args',
            nargs='*',
            help='additional command args',
        )
        ParserBuilder._add_limit_argument(parser)

    def _add_subparser(self, *args, **kwargs):
        subparser = self.subparsers.add_parser(*args, **kwargs)
        ParserBuilder._add_verbose_argument(subparser)
        return subparser

    def build(self):
        return self.parser

    @staticmethod
    def _add_verbose_argument(parser):
        parser.add_argument(
            '-v',
            '--verbose',
            default=0,
            action='count',
            help='be more verbose',
        )
        parser.add_argument(
            '-q',
            '--quite',
            default=0,
            action='count',
            help='be less verbose',
        )

    @staticmethod
    def _add_limit_argument(parser):
        parser.add_argument(
            '--limit',
            metavar='N',
            default=-1,
            type=int,
            help='limit duplicate results',
        )



import os
import logging

import sqlalchemy

from findd.services import Findd
from findd.utils.path import parents


__LOG__ = logging.getLogger(__name__)
DEBUG_MORE = 5


class ContextException(Exception):
    pass


class Context(object):

    def __init__(self, base_dir, findd_dir, db_url):
        assert base_dir is not None
        assert findd_dir is not None
        assert db_url is not None

        assert os.path.exists(base_dir)

        self.base_dir = base_dir
        self.findd_dir = findd_dir
        self.db_url = db_url

    def assert_findd_dir_does_not_exists(self):
        if os.path.exists(self.findd_dir):
            raise ContextException('.findd directory already exists')

    def assert_findd_dir_exists(self):
        if not os.path.exists(self.findd_dir):
            raise ContextException('.findd directory not found')

    def db_engine(self):
        echo = __LOG__.isEnabledFor(DEBUG_MORE)
        return sqlalchemy.create_engine(self.db_url, echo=echo)

    def db_sessionmaker(self):
        return sqlalchemy.orm.sessionmaker(bind=self.db_engine())

    def db_session(self):
        return self.db_sessionmaker()()

    def findd(self):
        return Findd(base_dir=self.base_dir, db_session=self.db_session())

    @property
    def is_excluded(self):
        return lambda path: os.path.basename(path) == '.findd'


def search_base_dir(start_dir=None):
    if start_dir is None:
        start_dir = os.getcwd()

    for path in parents(start_dir):
        if os.path.exists(os.path.join(path, '.findd')):
            return path

    return os.getcwd()


def create_from_base_dir(base_dir=None):
    if base_dir is None:
        base_dir = search_base_dir()

    findd_dir = os.path.join(base_dir, '.findd')
    db_path = os.path.join(findd_dir, 'findd.sqlite3')
    db_url = 'sqlite:///%s' % db_path

    return Context(
        base_dir=base_dir,
        findd_dir=findd_dir,
        db_url=db_url,
    )


def configure_logging(verbosity):
    levels = [
        logging.CRITICAL,
        logging.ERROR,
        logging.WARNING,
        logging.INFO,
        logging.DEBUG,
        DEBUG_MORE,
    ]
    logging.basicConfig(
        format='%(levelname)s: %(message)s',
        level=levels[2 + max(min(verbosity, 3), -2)]
    )
    logging.addLevelName(DEBUG_MORE, 'DEBUG')



from sqlalchemy import Column
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import func
from sqlalchemy.ext.declarative import declarative_base


BASE = declarative_base()


class File(BASE):
    __tablename__ = 'file'

    path = Column(String, primary_key=True)
    mtime = Column(Integer)
    size = Column(Integer)
    md5 = Column(String(32))
    sha1 = Column(String(40))
    sha224 = Column(String(56))
    sha256 = Column(String(64))
    sha384 = Column(String(96))
    sha512 = Column(String(128))

    def probably_equal(self, afile):
        if afile is None:
            return False
        return (self.size == afile.size and
                self.sha512 == afile.sha512 and
                self.sha384 == afile.sha384 and
                self.sha256 == afile.sha256 and
                self.sha224 == afile.sha224 and
                self.sha1 == afile.sha1 and
                self.md5 == afile.md5)

    def __setitem__(self, k, v):
        setattr(self, k, v)

    def update(self, dict_):
        for key, value in dict_.items():
            self[key] = value

Index(
    'idx_duplicates',
    File.size.desc(),
    File.sha512,
    File.sha384,
    File.sha256,
    File.sha224,
    File.sha1,
    File.md5,
    File.path,    # covering idx
    File.mtime,   # covering idx
    unique=True,  # covering idx
)


def create_schema(connectable):
    BASE.metadata.create_all(connectable)


class FileRegistry(object):
    def __init__(self, session):
        """
        :type session: sqlalchemy.orm.session.Session
        """
        self.session = session

    def find_duplicates(self, limit=-1):
        if limit == 0:
            return
        query = self.session.query(File).order_by(
            File.size.desc(),
            File.sha512,
            File.sha384,
            File.sha256,
            File.sha224,
            File.sha1,
            File.md5,
            File.path,
        )
        duplicates = []
        pivot_element = None
        index = 0
        for afile in query:
            if afile.probably_equal(pivot_element):
                duplicates.append(afile)
                continue
            if len(duplicates) > 1:
                yield duplicates
                index = index + 1
                if index == limit:
                    return
            pivot_element = afile
            duplicates = [afile]
        if len(duplicates) > 1:
            yield duplicates

    def count(self):
        return self.session.query(func.count(File.path)).scalar()

    def delete(self, db_file):
        self.session.delete(db_file)

    def find_all(self):
        return self.session.query(File)

    def add(self, entity):
        self.session.add(entity)



import hashlib


def hashfile(path, blocksize=65536):
    md5 = hashlib.md5()
    sha1 = hashlib.sha1()
    sha224 = hashlib.sha224()
    sha256 = hashlib.sha256()
    sha384 = hashlib.sha384()
    sha512 = hashlib.sha512()

    with open(path, 'rb') as afile:
        buf = afile.read(blocksize)

        while len(buf) > 0:
            md5.update(buf)
            sha1.update(buf)
            sha224.update(buf)
            sha256.update(buf)
            sha384.update(buf)
            sha512.update(buf)
            buf = afile.read(blocksize)

    return {
        'md5': md5.hexdigest(),
        'sha1': sha1.hexdigest(),
        'sha224': sha224.hexdigest(),
        'sha256': sha256.hexdigest(),
        'sha384': sha384.hexdigest(),
        'sha512': sha512.hexdigest(),
    }



"""find duplicate files

Setup script for findd.

usage:
    python setup.py install

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
https://packaging.python.org/en/latest/single_source_version.html
"""

import setuptools

from os import path
from setuptools.command.test import test as TestCommand # NOQA

about = {}
with open(path.join(path.dirname(__file__), "findd", "__about__.py")) as f:
    exec(f.read(), about)

with open('README.rst') as f:
    README_LINES = f.read().splitlines(True)

setuptools.setup(
    name='findd',
    version=about['__version__'],
    description=README_LINES[3].strip(),
    long_description=''.join(README_LINES[5:]).strip(),
    url=about['__url__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    license=about['__license__'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Archiving :: Backup',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    platforms=['any'],
    keywords='find duplicates cli database-assisted',
    packages=setuptools.find_packages(),
    install_requires=[
        'backports.shutil_get_terminal_size',
        'blinker',
        'inflection',
        'progressbar2',
        'shellescape',
        'sqlalchemy',
    ],
    entry_points={
        'console_scripts': [
            'findd=findd.cli.__main__:main'
        ],
    },
)



from gettext import gettext
import logging

from findd.cli import context
from findd.cli.parser import build as build_parser


def main(args=None):
    parser = build_parser()
    try:
        opts = parser.parse_args(args)
        if not hasattr(opts, 'func'):
            parser.error(gettext('too few arguments'))
    except SystemExit as err:
        return 2 if err.code is None else err.code

    context.configure_logging(opts.verbosity)

    try:
        opts.func(opts)
        return 0
    except Exception as err:
        if logging.getLogger().isEnabledFor(logging.DEBUG):
            raise

        err_module = getattr(err, '__module__', '')
        if err_module.split('.')[0] == 'findd':
            logging.error(err)
            return 1

        raise  # pragma: no cover



from unittest import TestCase

from findd.utils.format import sizeof_fmt


class SizeofFmt(TestCase):
    def test_should_format_bytes_correctly(self):
        self.assertEqual(sizeof_fmt(0), '0.0B')
        self.assertEqual(sizeof_fmt(5), '5.0B')

    def test_should_format_gigabytes_correctly(self):
        self.assertEqual(sizeof_fmt(7.8 * 1024 ** 3), '7.8GiB')

    def test_should_format_yotabytes_correctly(self):
        self.assertEqual(sizeof_fmt(42 * 1024 ** 8), '42.0YiB')



import errno
import logging
import sys

import findd.cli


if sys.version_info >= (3,):
    import signal
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def main(args=None):
    try:
        exit_code = findd.cli.main(args)

        # py2: "lost sys.stderr" workaround:
        for flush in [sys.stdout.flush, sys.stderr.flush]:
            try:
                flush()
            except:
                pass

        return exit_code
    except KeyboardInterrupt:
        if logging.getLogger().isEnabledFor(logging.DEBUG):
            raise
    except IOError as err:
        if err.errno == errno.EPIPE:
            return 0
        raise


if __name__ == "__main__":
    sys.exit(main())



from os import pathsep
from os.path import basename
from os.path import dirname
from os.path import join
from mock import patch
from mock import Mock
from unittest import TestCase

import findd.utils.path as sut


class Parents(TestCase):

    def test_should_work_with_empty_paths(self):
        self.assertEqual(list(sut.parents('')), [''])

    def test_should_work_with_posix_paths(self):
        ps = sut.parents('a/b/c/d')
        prev = next(ps)
        self.assertEqual(prev, 'a/b/c/d')
        if pathsep == '/':
            prev = next(ps)
            self.assertEqual(prev, 'a/b/c')
        for p in ps:
            self.assertEqual(p, dirname(prev))
            prev = p

    def test_should_work_with_nt_paths(self):
        ps = sut.parents('C:\\a\\b\\c\\d')
        prev = next(ps)
        self.assertEqual(prev, 'C:\\a\\b\\c\\d')
        if pathsep == '\\':
            prev = next(ps)
            self.assertEqual(prev, 'C:\\a\\b\\c')
        for p in ps:
            self.assertEqual(p, dirname(prev))
            prev = p


class FilesOfDir(TestCase):

    @patch('findd.utils.path.walk')
    def test_should_work_with_empty_dirs(self, walk):
        walk.return_value = [('virtual', [], [])]
        self.assertEqual(list(sut.files_of_dir('virtual')), [])
        self.assertEqual(len(walk.mock_calls), 1)

    @patch('findd.utils.path.walk')
    @patch('findd.utils.path.stat')
    def test_should_work_with_non_empty_dirs(self, stat, walk):
        walk.return_value = [
            ('virtual', ['dir'], []),
            ('virtual/dir', [], ['123.file'])
        ]
        path123 = 'virtual/dir/123.file'

        files = list(sut.files_of_dir('virtual'))
        self.assertEqual(len(stat.mock_calls), 1)
        self.assertEqual(len(walk.mock_calls), 1)
        stat.assert_called_once_with(path123)
        self.assertEqual(list(map(lambda e: e.path, files)), [path123])

    @patch('findd.utils.path.walk')
    @patch('findd.utils.path.stat')
    def test_should_respect_is_excluded(self, stat, walk):
        def exclude_dotfindd(path):
            return basename(path) == '.findd'
        walk.return_value = [
            ('virtual', ['.findd'], ['root.file']),
        ]
        is_excluded = Mock(side_effect=exclude_dotfindd)
        files = list(sut.files_of_dir('virtual', is_excluded=is_excluded))
        self.assertEqual(len(walk.mock_calls), 1)
        self.assertEqual(len(is_excluded.mock_calls), 2)
        self.assertEqual(len(stat.mock_calls), 1)
        self.assertEqual(walk.return_value[0][1], [])
        paths = list(map(lambda e: e.path, files))
        self.assertEqual(paths, [join('virtual', 'root.file')])



import sys


if sys.version_info < (3,):
    import codecs

    def u(string):
        return codecs.decode(string, 'utf8', 'replace')

    def s(unicode_string):
        return codecs.encode(unicode_string, 'utf8', 'replace')

else:
    def u(unicode_string):
        return unicode_string

    def s(unicode_string):
        return unicode_string
